import os

from sanic import Blueprint
from sanic.exceptions import SanicException
from sanic.response import json
from sanic_ext import openapi, validate

from ...cfg import PipelineConfig

bp = Blueprint("api_flowerpower_cfg", url_prefix="api/cfg")


@bp.get("/project")
async def get_project(request) -> json:
    cfg = request.app.ctx.pipeline_manager.cfg.project.to_dict()
    # cfg.pop("fs")
    return json({"cfg": cfg})


@bp.get("/pipeline/<pipeline_name>")
async def get_pipeline(request, pipeline_name) -> json:
    if pipeline_name != request.app.ctx.pipeline_manager.cfg.pipeline.name:
        request.app.ctx.pipeline_manager.load_config(pipeline_name)
    cfg = request.app.ctx.pipeline_manager.cfg.pipeline.to_dict()
    return json({"cfg": cfg})


@bp.post("/pipeline/<pipeline_name>")
@openapi.body({"application/json": PipelineConfig}, required=True)
@validate(json=PipelineConfig)
async def update_pipeline(request, pipeline_name, body: PipelineConfig) -> json:
    data = request.json
    if pipeline_name != request.app.ctx.pipeline_manager.cfg.pipeline.name:
        request.app.ctx.pipeline_manager.load_config(pipeline_name)
    cfg = request.app.ctx.pipeline_manager.cfg.pipeline.copy()
    cfg.update(data)
    try:
        cfg.to_yaml(
            os.path.join(
                "pipelines",
                pipeline_name + ".yml",
            ),
            fs=request.app.ctx.pipeline_manager.cfg.fs,
        )
    except NotImplementedError as e:
        raise SanicException(f"Update failed. {e}", status_code=404)
    cfg
    return json({"cfg": cfg})



from orjson import dumps, loads
from sanic import Sanic
from sanic.response import json

from .setup import setup


def create_app(args):
    # Note: args is the parsed command line arguments when the server is
    # started using sanic build-in server. e.g.
    # sanic src.flowerpower.http.main:create_app -factory --base-dir /path/to/base_dir
    # app = setup(args.base_dir)
    base_dir = args.base_dir if hasattr(args, "base_dir") else None
    storage_options = (
        eval(args.storage_options) if hasattr(args, "storage_options") else {}
    )
    app = Sanic("flowerpower", dumps=dumps, loads=loads)
    setup(app=app, base_dir=base_dir, storage_options=storage_options)

    @app.get("/")
    async def index(request):
        return json({"status": "success", "message": "Welcome to FlowerPower"})

    return app



from sanic import Sanic

from ..pipeline import PipelineManager
from ..scheduler import SchedulerManager
from .api.cfg import bp as bp_api_cfg
from .api.pipeline import bp as bp_api_pipeline
from .api.scheduler import bp as bp_api_scheduler


def setup(app: Sanic, base_dir: str | None = None, storage_options: dict | None = None):
    app.config.BASE_DIR = base_dir
    app.config.STORAGE_OPTIONS = storage_options

    @app.listener("before_server_start")
    def init_scheduler(app, loop):
        app.ctx.pipeline_manager = PipelineManager(
            base_dir=base_dir, storage_options=storage_options
        )
        app.ctx.scheduler = SchedulerManager(
            base_dir=base_dir, storage_options=storage_options
        )
        app.ctx.scheduler.start_worker(background=True)
        # app.ctx.abc = 123

    @app.listener("before_server_stop")
    def cleanup_scheduler(app, loop):
        if hasattr(app.ctx, "scheduler"):
            app.ctx.scheduler.stop_worker()

    app.blueprint(bp_api_pipeline)
    app.blueprint(bp_api_scheduler)
    app.blueprint(bp_api_cfg)

    return app



import os
from typing import Any

import datafusion
import duckdb
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
from fsspec import AbstractFileSystem
from fsspec.utils import get_protocol
from pydantic import BaseModel, ConfigDict, Field

from ..helpers.filesystem import get_filesystem
from ..helpers.polars import pl
from ..helpers.sql import sql2polars_filter, sql2pyarrow_filter
from ..helpers.storage_options import (AwsStorageOptions, AzureStorageOptions,
                                       GcsStorageOptions, GitHubStorageOptions,
                                       GitLabStorageOptions)


class BaseFileIO(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    path: str | list[str]
    storage_options: (
        AwsStorageOptions
        | GcsStorageOptions
        | AzureStorageOptions
        | GitHubStorageOptions
        | GitLabStorageOptions
        | dict[str, Any]
        | None
    ) = None
    fs: AbstractFileSystem | None = Field(default=None)
    format: str | None = None

    def model_post_init(self, __context):
        # self._update_storage_options_from_env()
        if isinstance(self.path, str):
            self.path = (
                self.path.replace(f"**/*.{format}", "")
                .replace("**", "")
                .replace("*", "")
                .rstrip("//")
                .rstrip("/")
            )

        self._update_fs()

    def _update_storage_options_from_aws_credentials(
        self,
        profile: str = "default",
        allow_invalid_certificates: bool = False,
        allow_http: bool = False,
    ) -> AwsStorageOptions:
        if isinstance(self.storage_options, AwsStorageOptions):
            self.storage_options = self.storage_options.to_aws_credentials(
                profile=profile,
                allow_invalid_certificates=allow_invalid_certificates,
                allow_http=allow_http,
            )
        self._update_fs()

    def _update_storage_options_from_env(
        self,
    ):
        protocol = get_protocol(self.path)
        if protocol == "s3":
            self.storage_options = AwsStorageOptions.from_env()
        elif protocol == "gs" or protocol == "gcs":
            self.storage_options = GcsStorageOptions.from_env()
        elif protocol == "az" or protocol == "abfs":
            self.storage_options = AzureStorageOptions.from_env()
        elif protocol == "github":
            self.storage_options = GitHubStorageOptions.from_env()
        elif protocol == "gitlab":
            self.storage_options = GitLabStorageOptions.from_env()
        self._update_fs()

    def _update_storage_options(self, **kwargs):
        self.storage_options = self.storage_options.model_copy(update=kwargs)
        self._update_fs()

    def _update_fs(self):
        self.fs = get_filesystem(
            self.path if isinstance(self.path, str) else self.path[0],
            self.storage_options,
            dirfs=False,
        )

    def _glob_path(self):
        if isinstance(self.path, list):
            return
        else:
            if self.format is not None:
                if f"**/*.{format}" in self.path:
                    return self.path
                else:
                    return f"{self.path}/**/*.{format}"
            else:
                if "**" in self.path:
                    return self.path
                else:
                    return f"{self.path}/**"

    def list_files(self):
        if isinstance(self.path, list):
            return self.path

        glob_path = self._glob_path()
        return self.fs.glob(glob_path)


class BaseFileLoader(BaseFileIO):
    include_file_path: bool = False
    concat: bool = True
    conn: duckdb.DuckDBPyConnection | None = None
    ctx: datafusion.SessionContext | None = None

    def _load(self, **kwargs):
        if "include_file_path" not in kwargs:
            self.include_file_path = kwargs.get(
                "include_file_path", self.include_file_path
            )
        if "concat" not in kwargs:
            self.concat = kwargs.get("concat", self.concat)

        if not hasattr(self, "_data"):
            self._data = self.fs.read_files(
                self._glob_path,
                self.format,
                include_file_path=self.include_file_path,
                concat=self.concat,
                jsonlines=self.jsonlines or None,
                **kwargs,
            )
        else:
            reload = False
            if self.include_file_path and not "file_path" not in self._data.columns:
                reload = True
            if isinstance(self._data, pl.DataFrame) and not self.concat:
                reload = True
            if isinstance(self._data, list) and self.concat:
                self._data = pl.concat(self._data, how="diagonal_relaxed")
            if reload:
                self._load(
                    include_file_path=self.include_file_path,
                    concat=self.concat,
                    jsonlines=self.jsonlines or None,
                    **kwargs,
                )

    def to_pandas(self, **kwargs) -> pd.DataFrame | list[pd.DataFrame]:
        self._load(**kwargs)
        if not self.concat:
            return [df.to_pandas() for df in self._data]
        return self._data.to_pandas()

    def _to_polars_dataframe(self, **kwargs) -> pl.DataFrame | list[pl.DataFrame]:
        self._load(self, **kwargs)
        if not self.concat:
            return [
                df if isinstance(self._data, pl.DataFrame) else pl.from_arrow(df)
                for df in self._data
            ]
        return (
            self._data
            if isinstance(self._data, pl.DataFrame)
            else pl.from_arrow(self._data)
        )

    def _to_polars_lazyframe(self, **kwargs) -> pl.LazyFrame | list[pl.LazyFrame]:
        self._load(**kwargs)
        if not self.concat:
            return [df.lazy() for df in self._to_polars_dataframe()]
        return self._to_polars_dataframe.lazy()

    def to_polars(
        self,
        lazy: bool = False,
        **kwargs,
    ) -> pl.DataFrame | pl.LazyFrame | list[pl.DataFrame] | list[pl.LazyFrame]:
        if lazy:
            return self._to_polars_lazyframe(**kwargs)
        return self._to_polars_dataframe(**kwargs)

    def to_pyarrow_table(self, **kwargs) -> pa.Table | list[pa.Table]:
        self._load(**kwargs)
        if not self.concat:
            return [
                df.to_arrow(**kwargs) if isinstance(df, pl.DataFrame) else df
                for df in self._data
            ]
        return (
            self._data.to_arrow(**kwargs)
            if isinstance(self._data, pl.DataFrame)
            else self._data
        )

    def to_duckdb_relation(
        self, conn: duckdb.DuckDBPyConnection | None = None, **kwargs
    ) -> duckdb.DuckDBPyRelation:
        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if hasattr(self, "_data"):
            return self.conn.from_arrow(self.to_pyarrow_table(concat=True))

        else:
            return self.conn.read_csv(self._glob_path, **kwargs)

    def register_in_duckdb(
        self, conn: duckdb.DuckDBPyConnection, name: str | None = None, **kwargs
    ) -> None:
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if hasattr(self, "_data"):
            self.conn.register(name, self.to_pyarrow_table(concat=True))
        else:
            self.conn.register(name, self.to_duckdb_relation(self.conn, **kwargs))

    def register_in_datafusion(
        self, ctx: datafusion.SessionContext, name: str | None = None, **kwargs
    ) -> None:
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.ctx is None:
            if ctx is None:
                ctx = datafusion.SessionContext()
            self.ctx = ctx

        self.ctx.register_record_batches(
            name, [self.to_pyarrow_table(concat=True).to_batches()]
        )
        # return ctx

    def filter(
        self, filter_expr: str | pl.Expr | pa.compute.Expression
    ) -> (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | list[pl.DataFrame]
        | list[pl.LazyFrame]
        | list[pa.Table]
    ):
        if isinstance(self._data, pl.DataFrame | pl.LazyFrame):
            pl_schema = (
                self._data.schema
                if isinstance(self._data, pl.DataFrame)
                else self._data.collect_schema()
            )
            filter_expr = (
                sql2polars_filter(filter_expr, pl_schema)
                if isinstance(filter_expr, str)
                else filter_expr
            )
            return self._data.filter(filter_expr)

        elif isinstance(self._data, pa.Table):
            pa_schema = self._data.schema
            filter_expr = (
                sql2pyarrow_filter(filter_expr, pa_schema)
                if isinstance(filter_expr, str)
                else filter_expr
            )
            return self._data.filter(filter_expr)

        if isinstance(self._data, str):
            if isinstance(self._data[0], pl.DataFrame | pl.LazyFrame):
                pl_schema = (
                    self._data.schema
                    if isinstance(self._data[0], pl.DataFrame)
                    else self._data[0].collect_schema()
                )
                filter_expr = (
                    sql2polars_filter(filter_expr, pl_schema)
                    if isinstance(filter_expr, str)
                    else filter_expr
                )
                return [d.filter(filter_expr) for d in self._data]
            elif isinstance(self._data[0], pa.Table):
                pa_schema = self._data[0].schema
                filter_expr = (
                    sql2pyarrow_filter(filter_expr, pa_schema)
                    if isinstance(filter_expr, str)
                    else filter_expr
                )
                return [d.filter(filter_expr) for d in self._data]


class BaseDatasetLoader(BaseFileLoader):
    _schema: pa.Schema | None = None
    partitioning: str | list[str] | pds.Partitioning | None = None

    def to_pyarrow_dataset(
        self,
        **kwargs,
    ) -> pds.Dataset:
        if self.format == ["csv", "arrow", "ipc"]:
            self._dataset = self.fs.pyarrow_dataset(
                self.path,
                format=self.format,
                schema=self._schema,
                partitioning=self.partitioning,
                **kwargs,
            )
        elif self.format == "parquet":
            if self.fs.exists(os.path.join(self.path, "_metadata")):
                self._dataset = self.fs.parquet_dataset(
                    self.path,
                    schema=self._schema,
                    partitioning=self.partitioning,
                    **kwargs,
                )
            else:
                self._dataset = self.fs.pyarrow_dataset(
                    self.path,
                    format=self.format,
                    schema=self._schema,
                    partitioning=self.partitioning,
                    **kwargs,
                )
        else:
            self._dataset = pds.dataset(
                self.to_pyarrow_table(**kwargs), schema=self._schema
            )

    def to_pandas(self, **kwargs) -> pd.DataFrame:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return self._dataset.to_table().to_pandas()

    def _to_polars_dataframe(self, **kwargs) -> pl.DataFrame:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return pl.from_arrow(self._dataset.to_table())

    def _to_polars_lazyframe(self, **kwargs) -> pl.LazyFrame:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return pl.scan_pyarrow_dataset(self._dataset)

    def to_polars(self, lazy: bool = True, **kwargs) -> pl.DataFrame | pl.LazyFrame:
        return (
            self._to_polars_lazyframe(**kwargs)
            if lazy
            else self._to_polars_dataframe(**kwargs)
        )

    def to_pyarrow_table(self, **kwargs) -> pa.Table:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return self._dataset.to_table()

    def to_duckdb_relation(
        self, conn: duckdb.DuckDBPyConnection | None = None, **kwargs
    ) -> duckdb.DuckDBPyRelation:
        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)

        return self.conn.from_arrow(self._dataset)

    def register_in_duckdb(
        self, conn: duckdb.DuckDBPyConnection, name: str | None = None, **kwargs
    ) -> None:
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)

        self.conn.register(name, self._dataset)
        # return conn

    def register_in_datafusion(
        self, ctx: datafusion.SessionContext, name: str | None = None, **kwargs
    ) -> None:
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.ctx is None:
            if ctx is None:
                ctx = datafusion.SessionContext()
            self.ctx = ctx

        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)

        self.ctx.register_dataset(name, self._dataset)
        # return self.tx

    def filter(self, filter_exp: str | pa.compute.Expression) -> pds.Dataset:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset()
        if isinstance(filter_exp, str):
            filter_exp = sql2pyarrow_filter(filter_exp, self._dataset.schema)
        return self._dataset.filter(filter_exp)


class BaseFileWriter(BaseFileIO):
    data: (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pd.DataFrame
        | list[pl.DataFrame]
        | list[pl.LazyFrame]
        | list[pa.Table]
        | list[pd.DataFrame]
    )
    mode: str = "append"

    def model_post_init(self, __context):
        if isinstance(self.data, list) and self.concat:
            if isinstance(self.data[0], pl.DataFrame | pl.LazyFrame):
                self.data = pl.concat(self.data, how="diagonal_relaxed")
            elif isinstance(self.data[0], pa.Table):
                self.data = pa.concat_tables(self.data, promote_options="permissive")
            elif isinstance(self.data[0], pd.DataFrame):
                self.data = pd.concat(self.data)
        self._gen_paths()

    def _gen_paths(self):
        if isinstance(self.path, str) and isinstance(self.data, list):
            if self.path.endswith(self.format):
                self.path = [
                    f"{self.path}-{i}.{self.format}" for i in range(len(self.data))
                ]

    def write(self, data, **kwargs):
        self.fs.write



import datetime as dt

import datafusion as dtf
import duckdb
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
from deltalake import DeltaTable
from hamilton.function_modifiers import dataloader

from ...helpers.polars import pl
from ..utils import get_dataframe_metadata, get_delta_metadata


@dataloader()
def load(
    path: str, storage_options: dict[str, str] | None = None, **kwargs
) -> tuple[DeltaTable, dict]:
    dtable = DeltaTable(path, storage_options=storage_options, **kwargs)
    metadata = get_delta_metadata(dtable, path, **kwargs)

    return dtable, metadata


@dataloader()
def load_pyarrow_table(
    path: str, storage_options: dict[str, str] | None = None, **kwargs
) -> tuple[pa.Table, dict]:
    dtable, metadata = load(path, storage_options=storage_options, **kwargs)
    table = dtable.to_pyarrow_table()
    metadata["num_rows"] = table.num_rows
    return table, metadata


@dataloader()
def load_pyarrow_dataset(
    path: str, storage_options: dict[str, str] | None = None, **kwargs
) -> dict[pds.Dataset, dict]:
    dtable, metadata = load(path, storage_options=storage_options, **kwargs)
    ds = dtable.to_pyarrow_dataset()
    return ds, metadata


@dataloader()
def load_pandas_dataframe(
    path: str, storage_options: dict[str, str] | None = None, **kwargs
) -> tuple[pd.DataFrame, dict]:
    dtable, metadata = load(path, storage_options=storage_options, **kwargs)
    df = dtable.to_pandas()
    metadata["num_rows"] = df.shape[0]
    return df, metadata


@dataloader()
def load_polars_dataframe(
    path: str, storage_options: dict[str, str] | None = None, **kwargs
) -> tuple[pl.DataFrame, dict]:
    df = pl.read_delta(path, storage_options=storage_options, **kwargs)
    metadata = get_dataframe_metadata(df, path, **kwargs)

    return df, metadata


@dataloader()
def load_polars_lazyframe(
    path: str, storage_options: dict[str, str] | None = None, **kwargs
) -> pl.LazyFrame:
    df = pl.scan_delta(path, storage_options=storage_options, **kwargs)
    df_schema = df.collect_schema.to_python()
    metadata = {
        "path": path,
        "format": "delta",
        "timestamp": dt.datetime.now().timestamp(),
        "schema": df_schema,
        "partition_columns": None,
        "num_columns": len(df_schema),
        "num_rows": None,
        "name": None,
        "description": None,
        "id": None,
    }
    return df, metadata


@dataloader()
def load_duckdb_relation(
    path: str,
    storage_options: dict[str, str] | None = None,
    conn: duckdb.DuckDBPyConnection | None = None,
    **kwargs,
) -> tuple[duckdb.DuckDBPyRelation, dict]:
    ds, metadata = load_pyarrow_dataset(path, storage_options=storage_options, **kwargs)
    if conn is None:
        conn = duckdb.connect()

    table = conn.from_arrow(ds)

    return table, metadata


@dataloader()
def register_delta_in_duckdb(
    path: str,
    storage_options: dict[str, str] | None = None,
    name: str | None = None,
    conn: duckdb.DuckDBPyConnection | None = None,
    lazy: bool = True,
    **kwargs,
) -> tuple[duckdb.DuckDBPyConnection, dict]:
    if name is None:
        name = "delta_table"
    table, metadata = load_pyarrow_dataset(
        path, storage_options=storage_options, conn=conn, name=name, **kwargs
    )
    if not lazy:
        table = table.to_table()
    if conn is None:
        conn = duckdb.connect()
    conn.register(name, table)
    return conn, metadata


@dataloader()
def register_delta_in_datafusion(
    path: str,
    storage_options: dict[str, str] | None = None,
    name: str | None = None,
    ctx: dtf.SessionContext | None = None,
    lazy: bool = True,
    **kwargs,
) -> tuple[dtf.SessionContext, dict]:
    if name is None:
        name = "delta_table"
    table, metadata = (
        load_pyarrow_dataset(
            path, storage_options=storage_options, name=name, **kwargs
        ),
    )

    if ctx is None:
        ctx = dtf.SessionContext()

    if not lazy:
        ctx.register_record_batches(name, [table.to_batches()])
    else:
        ctx.register_dataset(name, table)

    return ctx, metadata



import random
import time
from typing import Callable

from loguru import logger
from munch import Munch
from paho.mqtt.client import CallbackAPIVersion, Client

from ..cfg import Config


class MQTTClient:
    def __init__(
        self,
        user: str | None = None,
        pw: str | None = None,
        host: str | None = "localhost",
        port: int | None = 1883,
        topic: str | None = None,
        first_reconnect_delay: int = 1,
        max_reconnect_count: int = 5,
        reconnect_rate: int = 2,
        max_reconnect_delay: int = 60,
        transport: str = "tcp",
    ):
        self.topic = topic

        self._user = user
        self._pw = pw
        self._host = host
        self._port = port
        self._first_reconnect_delay = first_reconnect_delay
        self._max_reconnect_count = max_reconnect_count
        self._reconnect_rate = reconnect_rate
        self._max_reconnect_delay = max_reconnect_delay
        self._transport = transport

        self._client = None

    def connect(self) -> Client:
        def on_connect(client, userdata, flags, rc, properties):
            if rc == 0:
                logger.info(f"Connected to MQTT Broker {userdata.host}!")
            else:
                logger.error(f"Failed to connect, return code {rc}")

        def on_disconnect(client, userdata, disconnect_flags, rc, properties=None):
            logger.info(f"Disconnected with result code: {rc}")
            reconnect_count, reconnect_delay = 0, userdata.first_reconnect_delay

            if userdata.max_reconnect_count == 0:
                logger.info("Disconnected successfully!")
                return

            while reconnect_count < userdata.max_reconnect_count:
                logger.info(f"Reconnecting in {reconnect_delay} seconds...")
                time.sleep(reconnect_delay)

                try:
                    client.reconnect()
                    logger.info("Reconnected successfully!")
                    return
                except Exception as err:
                    logger.error(f"{err}. Reconnect failed. Retrying...")

                reconnect_delay *= userdata.reconnect_rate
                reconnect_delay = min(reconnect_delay, userdata.max_reconnect_delay)
                reconnect_count += 1
            logger.info(
                f"Reconnect failed after {reconnect_count} attempts. Exiting..."
            )

        client = Client(
            CallbackAPIVersion.VERSION2,
            client_id=f"ie-databus-{random.randint(0, 10000)}",
            transport=self._transport,
            userdata=Munch(
                user=self._user,
                pw=self._pw,
                host=self._host,
                port=self._port,
                topic=self.topic,
                first_reconnect_delay=self._first_reconnect_delay,
                max_reconnect_count=self._max_reconnect_count,
                reconnect_rate=self._reconnect_rate,
                max_reconnect_delay=self._max_reconnect_delay,
                transport=self._transport,
            ),
        )
        if self._pw != "" and self._user != "":
            client.username_pw_set(self._user, self._pw)

        client.on_connect = on_connect
        client.on_disconnect = on_disconnect

        client.connect(self._host, self._port)

        # topic = topic or topic
        if self.topic:
            self.subscribe()

        self._client = client

    def disconnect(self):
        self._max_reconnect_count = 0
        self._client._userdata.max_reconnect_count = 0
        self._client.disconnect()

    def reconnect(self):
        self._client.reconnect()

    def publish(self, topic, payload):
        if self._client is None:
            self.connect()
        elif self._client.is_connected() is False:
            self.reconnect()
        self._client.publish(topic, payload)

    def subscribe(self, topic: str | None = None):
        if topic is not None:
            self.topic = topic
        self._client.subscribe(self.topic)

    def unsubscribe(self, topic: str | None = None):
        if topic is not None:
            self.topic = topic
        self._client.unsubscribe(self.topic)

    def register_on_message(self, on_message: Callable):
        self._client.on_message = on_message

    def run_in_background(
        self,
        on_message: Callable,
        topic: str | None = None,
    ) -> Client:
        if self._client is None or not self._client.is_connected():
            self.connect()

        if topic:
            self.subscribe(topic)

        self._client.on_message = on_message
        self._client.loop_start()

    def run_until_break(
        self,
        on_message: Callable,
        topic: str | None = None,
    ):
        if self._client is None or not self._client.is_connected():
            self.connect()

        if topic:
            self.subscribe(topic)

        self._client.on_message = on_message
        self._client.loop_forever()

    def start_listener(
        self, on_message: Callable, topic: str | None = None, background: bool = False
    ):
        if background:
            self.run_in_background(on_message, topic)
        else:
            self.run_until_break(on_message, topic)

    def stop_listener(
        self,
    ):
        self._client.loop_stop()
        logger.info("Client stopped.")

    @classmethod
    def from_event_broker(cls, base_dir: str):
        event_broker_cfg = Config.load(base_dir=base_dir).project.worker.event_broker
        if event_broker_cfg is not None:
            if event_broker_cfg.get("type", None) == "mqtt":
                return cls(
                    user=event_broker_cfg.get("user", None),
                    pw=event_broker_cfg.get("pw", None),
                    host=event_broker_cfg.get("host", "localhost"),
                    port=event_broker_cfg.get("port", 1883),
                    transport=event_broker_cfg.get("transport", "tcp"),
                )
            raise ValueError("No event broker configuration found in config file.")
        else:
            raise ValueError("No event broker configuration found in config file.")

    @classmethod
    def from_config(cls, cfg: dict):
        return cls(
            user=cfg.get("user", None),
            pw=cfg.get("pw", None),
            host=cfg.get("host", "localhost"),
            port=cfg.get("port", 1883),
            transport=cfg.get("transport", "tcp"),
        )



from sanic import Blueprint
from sanic.response import json, raw

bp = Blueprint("api_flowerpower_pipeline", url_prefix="api/pipeline")


@bp.post("add/<name>")
async def add_(
    request,
    name: str,
) -> json:
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.add(name=name, **kwargs)
    return json({"status": "success"})


@bp.post("/add-job/<name>")
async def add_job(
    request,
    name: str,
) -> json:
    kwargs = request.json or {}
    id_ = request.app.ctx.pipeline_manager.add_job(name=name, **kwargs)
    return json({"job_id": str(id_)})


@bp.get("/all-pipelines")
async def all_pipelines(
    request,
) -> json:
    pipelines = request.app.ctx.pipeline_manager.all_pipelines(show=False)
    return json({"pipelines": pipelines})


@bp.delete("/delete/<name>")
async def delete(
    request,
    name: str,
) -> json:
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.delete(name, **kwargs)
    return json({"status": "success"})


@bp.post("/new/<name>")
async def new(
    request,
    name: str,
) -> json:
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.new(name, **kwargs)
    return json({"status": "success"})


@bp.get("/summary")
async def summary_all(
    request,
    name: str | None = None,
) -> json:
    pipeline = request.app.ctx.pipeline_manager.get_summary(name)
    return json({"pipeline": pipeline})


@bp.get("/summary/<name>")
async def summary(
    request,
    name: str | None = None,
) -> json:
    pipeline = request.app.ctx.pipeline_manager.get_summary(name)
    return json({"pipeline": pipeline})


@bp.post("/run/<name>")
async def run_(
    request,
    name: str,
) -> json:
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.run(name, **kwargs)
    return json({"status": "success"})


@bp.post("/run-job/<name>")
async def run_job(
    request,
    name: str,
) -> json:
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.run_job(name, **kwargs)
    return json({"status": "success"})


@bp.post("/schedule/<name>")
async def schedule_pipeline(request, name):
    kwargs = request.json or {}
    id_ = request.app.ctx.pipeline_manager.schedule(name, **kwargs)
    return json({"schedule_id": str(id_)})


@bp.post("/start-mqtt-listener/<name>")
async def start_mqtt_listener(request, name):
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.start_mqtt_listener(name, **kwargs)
    return json({"status": "success"})


@bp.post("/stop-mqtt-listener/<name>")
async def stop_mqtt_listener(request, name):
    kwargs = request.json or {}
    _ = request.app.ctx.pipeline_manager.stop_mqtt_listener(name, **kwargs)
    return json({"status": "success"})


@bp.get("/show/<name>")
async def show(
    request,
    name: str,
) -> json:
    pipeline_dag = request.app.ctx.pipeline_manager.show_dag(name=name)
    return raw(pipeline_dag.pipe("svg"))


@bp.post("/set-abc/<value>")
async def set_abc(request, value):
    request.app.ctx.abc = value
    return json({"status": "success"})



import os

import pyarrow as pa
import tqdm
from joblib import Parallel, delayed


def convert_large_types_to_standard(schema: pa.Schema) -> pa.Schema:
    # Define mapping of large types to standard types
    type_mapping = {
        pa.large_string(): pa.string(),
        pa.large_binary(): pa.binary(),
        pa.large_list(pa.null()): pa.list_(pa.null()),
    }

    # Convert fields
    new_fields = []
    for field in schema:
        field_type = field.type
        # Check if type exists in mapping
        if field_type in type_mapping:
            new_field = pa.field(
                name=field.name,
                type=type_mapping[field_type],
                nullable=field.nullable,
                metadata=field.metadata,
            )
            new_fields.append(new_field)
        # Handle large lists with nested types
        elif isinstance(field_type, pa.LargeListType):
            new_field = pa.field(
                name=field.name,
                type=pa.list_(field_type.value_type),
                nullable=field.nullable,
                metadata=field.metadata,
            )
            new_fields.append(new_field)
        else:
            new_fields.append(field)

    return pa.schema(new_fields)


def run_parallel(
    func: callable,
    func_params: list[any],
    *args,
    n_jobs: int = -1,
    backend: str = "threading",
    verbose: bool = True,
    **kwargs,
) -> list[any]:
    """Runs a function for a list of parameters in parallel.

    Args:
        func (Callable): function to run in Parallelallel.
        func_params (list[any]): parameters for the function
        n_jobs (int, optional): Number of joblib workers. Defaults to -1.
        backend (str, optional): joblib backend. Valid options are
        `loky`,`threading`, `mutliprocessing` or `sequential`.  Defaults to "threading".

    Returns:
        list[any]: Function output.
    """
    if verbose:
        return Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(func)(fp, *args, **kwargs) for fp in tqdm.tqdm(func_params)
        )

    else:
        return Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(func)(fp, *args, **kwargs) for fp in func_params
        )


def get_partitions_from_path(
    path: str, partitioning: str | list[str] | None = None
) -> list[tuple]:
    """Get the dataset partitions from the file path.

    Args:
        path (str): File path.
        partitioning (str | list[str] | None, optional): Partitioning type. Defaults to None.

    Returns:
        list[tuple]: Partitions.
    """
    if "." in path:
        path = os.path.dirname(path)

    parts = path.split("/")

    if isinstance(partitioning, str):
        if partitioning == "hive":
            return [tuple(p.split("=")) for p in parts if "=" in p]

        else:
            return [
                (partitioning, parts[0]),
            ]
    else:
        return list(zip(partitioning, parts[-len(partitioning) :]))



import datetime as dt
import os

import duckdb
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as pds
from hamilton.function_modifiers import dataloader
from pydantic import BaseModel


class DuckDBLoader(BaseModel):
    path: str | None = None
    read_only: bool = False
    conn: duckdb.DuckDBPyConnection | None = None
    host: str | None = None
    port: int | None = None
    socket: str | None = None
    user: str | None = None
    password: str | None = None
    database: str | None = None

    def model_post_init(self, __context):
        if self.path is None:
            self.path = ":memory:"
        if self.conn is None:
            self.conn = duckdb.connect(self.path, self.read_only)

        if self.host is not None:
            self._update_from_env()

        if self.host is not None:
            if self.port == 5432:
                self.attach_postgres()
            elif self.port == 3306:
                self.attach_mysql()

    def _update_from_env(self):
        self.host = (
            os.getenv("POSTGRES_HOST") or os.getenv("PGHOST") or os.getenv("MYSQL_HOST")
        )
        self.port = (
            os.getenv("POSTGRES_PORT")
            or os.getenv("PGPORT")
            or os.getenv("MYSQL_TCP_PORT")
        )
        self.user = (
            os.getenv("POSTGRES_USER") or os.getenv("PGUSER") or os.getenv("MYSQL_USER")
        )
        self.password = (
            os.getenv("POSTGRES_PASSWORD")
            or os.getenv("PGPASSWORD")
            or os.getenv("MYSQL_PW")
        )
        self.database = (
            os.getenv("POSTGRES_DB")
            or os.getenv("PGDATABASE")
            or os.getenv("MYSQL_DATABASE")
        )
        self.socket = os.getenv("MYSQL_UNIX_PORT")

    def attach_postgres(
        self,
    ):
        self.conn.execute("INSTALL postgres; LOAD postgres;")
        connection_string = ""
        for k, v in {
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password,
            "database": self.database,
        }.items():
            if v is not None:
                connection_string += f"{k}={v} "

        self.conn.execute(
            f"ATTACH {connection_string} AS postgres_db (TYPE POSTGRES, READ_ONLY);"
        )

    def attach_mysql(
        self,
    ):
        self.conn.execute("INSTALL mysql; LOAD mysql;")
        connection_string = ""
        for k, v in {
            "host": self.host,
            "port": self.port,
            "socket": self.socket,
            "user": self.user,
            "password": self.password,
            "database": self.database,
        }.items():
            if v is not None:
                connection_string += f"{k}={v} "
        self.conn.execute(
            f"ATTACH {connection_string} AS mysql_db (TYPE MYSQL, READ_ONLY);"
        )

    def to_duckdb_relation(self, query: str) -> duckdb.DuckDBPyRelation:
        return self.conn.sql(query)

    def to_pandas(self, query: str) -> pd.DataFrame:
        return self.to_duckdb_relation(query).tdf()

    def to_polars(self, query: str) -> pl.DataFrame:
        return self.to_duckdb_relation(query).pl()

    def to_pyarrow_table(self, query: str) -> pa.Table:
        return self.to_duckdb_relation(query).arrow()

    def to_dataset(self, query: str) -> pds.Dataset:
        return pds.dataset(self.to_pyarrow_table(query))


@dataloader()
def load_from_duckdb(
    query: str,
    path: str | None = None,
    conn: duckdb.DuckDBPyConnection | None = None,
    **kwargs,
) -> tuple[duckdb.DuckDBPyRelation, dict]:
    """
    Load data from a DuckDB database.

    Args:
        query: (str) SQL query.
        path: (str) Path to the database file. If None, an in-memory database is created.
        conn: (DuckDBPyConnection) DuckDB connection.
        **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

    Returns:
        table: (DuckDBPyRelation) DuckDB relation.
        metadata: (dict) Metadata dictionary.
    """
    if conn is None:
        conn = duckdb.connect(path, **kwargs)

    table = conn.sql(query)
    metadata = {
        "path": path,
        "format": "duckdb",
        "timestamp": dt.datetime.now().timestamp(),
        "query": query,
    }
    return table, metadata


@dataloader()
def load_from_sqlite(
    query: str,
    path: str | None = None,
    conn: duckdb.DuckDBPyConnection | None = None,
    **kwargs,
) -> tuple[duckdb.DuckDBPyRelation, dict]:
    """
    Load data from a SQLite database.

    Args:
        query: (str) SQL query.
        path: (str) Path to the database file. If None, an in-memory database is created.
        conn: (DuckDBPyConnection) DuckDB connection.
        **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

    Returns:
        table: (DuckDBPyRelation) DuckDB relation.
        metadata: (dict) Metadata dictionary.
    """
    if conn is None:
        conn = duckdb.connect(path, **kwargs)

    table = conn.sql(query)
    metadata = {
        "path": path,
        "format": "sqlite",
        "timestamp": dt.datetime.now().timestamp(),
        "query": query,
    }
    return table, metadata


@dataloader()
def load_from_postgres(
    query: str,
    path: str | None = None,
    conn: duckdb.DuckDBPyConnection | None = None,
    host: str | None = None,
    port: int | None = None,
    user: str | None = None,
    password: str | None = None,
    database: str | None = None,
    **kwargs,
) -> tuple[duckdb.DuckDBPyRelation, dict]:
    """
    Load data from a PostgreSQL database.

    Args:
        query: (str) SQL query.
        path: (str) Path to the database file. If None, an in-memory database is created.
        conn: (DuckDBPyConnection) DuckDB connection.
        host: (str) PostgreSQL host.
        port: (int) PostgreSQL port.
        user: (str) PostgreSQL user.
        password: (str) PostgreSQL password.
        database: (str) PostgreSQL database.
        **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

    Returns:
        table: (DuckDBPyRelation) DuckDB relation.
        metadata: (dict) Metadata dictionary.
    """
    if conn is None:
        if path is None:
            path = ":memory:"

        conn = duckdb.connect(path, **kwargs)

    conn.execute("INSTALL postgres; LOAD postgres;")
    connection_string = " ".join(
        f"{k}={v}"
        for k, v in {
            "host": host,
            "port": port,
            "user": user,
            "password": password,
            "database": database,
        }.items()
        if v is not None
    )
    conn.execute(
        f"ATTACH {connection_string} AS postgres_db (TYPE POSTGRES, READ_ONLY);"
    )

    table = conn.sql(query)
    metadata = {
        "path": path,
        "format": "postgres",
        "timestamp": dt.datetime.now().timestamp(),
        "query": query,
    }
    return table, metadata


@dataloader()
def load_from_mysql(
    query: str,
    path: str | None = None,
    conn: duckdb.DuckDBPyConnection | None = None,
    host: str | None = None,
    port: int | None = None,
    user: str | None = None,
    password: str | None = None,
    database: str | None = None,
    **kwargs,
) -> tuple[duckdb.DuckDBPyRelation, dict]:
    """
    Load data from a MySQL database.

    Args:
        query: (str) SQL query.
        path: (str) Path to the database file. If None, an in-memory database is created.
        conn: (DuckDBPyConnection) DuckDB connection.
        host: (str) MySQL host.
        port: (int) MySQL port.
        user: (str) MySQL user.
        password: (str) MySQL password.
        database: (str) MySQL database.
        **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

    """
    if conn is None:
        if path is None:
            path = ":memory:"

        conn = duckdb.connect(path, **kwargs)

    conn.execute("INSTALL mysql; LOAD mysql;")
    connection_string = " ".join(
        f"{k}={v}"
        for k, v in {
            "host": host,
            "port": port,
            "user": user,
            "password": password,
            "database": database,
        }.items()
        if v is not None
    )
    conn.execute(f"ATTACH {connection_string} AS mysql_db (TYPE MYSQL, READ_ONLY);")

    table = conn.sql(query)
    metadata = {
        "path": path,
        "format": "mysql",
        "timestamp": dt.datetime.now().timestamp(),
        "query": query,
    }
    return table, metadata



from ..base import BaseDatasetLoader, BaseFileLoader


class CSVFileLoader(BaseFileLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "csv"


class CSVDatasetLoader(BaseDatasetLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "csv"



import importlib

has_pyarrow = importlib.util.find_spec("pyarrow") is not None

if has_pyarrow:
    from .ext import AbstractFileSystem
else:
    from fsspec import AbstractFileSystem

from .base import get_filesystem



import uuid

import dill
from sanic import Blueprint
from sanic.exceptions import SanicException
from sanic.response import json, raw

bp = Blueprint("api_flowerpower_scheduler", url_prefix="api/scheduler")


@bp.get("/status")
async def status(request) -> json:
    status = request.app.ctx.scheduler.state
    return json({"status": status.__str__()})


@bp.get("/jobs")
async def jobs(request) -> json:
    jobs = request.app.ctx.scheduler.get_jobs(as_dict=True)
    return json({"jobs": jobs})


@bp.get("/job-result/<job_id>")
async def job_result(request, job_id) -> raw:
    job_id = uuid.UUID(job_id)
    if job_id not in [job.id for job in request.app.ctx.scheduler.get_jobs()]:
        raise SanicException("Job not found", status_code=404)
    job = request.app.ctx.scheduler.get_job_result(job_id)
    return raw(dill.dumps(job))


@bp.get("/schedules")
async def schedules(request) -> json:
    schedules = request.app.ctx.scheduler.get_schedules(as_dict=True)
    return json({"schedules": schedules})


@bp.get("/schedule/<schedule_id>")
async def schedule(request, schedule_id) -> json:
    if schedule_id not in [
        sched.id for sched in request.app.ctx.scheduler.get_schedules()
    ]:
        raise SanicException("Schedule not found", status_code=404)
    schedule = request.app.ctx.scheduler.get_schedule(schedule_id, as_dict=True)
    return json(schedule)


@bp.delete("/schedule/<schedule_id>")
async def remove_schedule(request, schedule_id) -> json:
    if schedule_id not in [
        sched.id for sched in request.app.ctx.scheduler.get_schedules()
    ]:
        raise SanicException("Schedule not found", status_code=404)
    request.app.ctx.scheduler.remove_schedule(schedule_id)
    return json({"status": "success"})


@bp.delete("/schedules")
async def remove_schedules(request) -> json:
    request.app.ctx.scheduler.remove_all_schedules()
    return json({"status": "success"})


@bp.post("/pause-schedule/<schedule_id>")
async def pause_schedule(request, schedule_id) -> json:
    if schedule_id not in [
        sched.id for sched in request.app.ctx.scheduler.get_schedules()
    ]:
        raise SanicException("Schedule not found", status_code=404)
    request.app.ctx.scheduler.pause_schedule(schedule_id)
    return json({"status": "success"})


@bp.get("/tasks")
async def tasks(request) -> json:
    tasks = request.app.ctx.scheduler.get_tasks(as_dict=True)
    return json({"tasks": tasks})



from ..base import BaseDatasetLoader, BaseFileLoader


class ParquetFileLoader(BaseFileLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "json"


class ParquetDatasetLoader(BaseDatasetLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "json"



import datetime as dt
import pathlib
from typing import Any

import yaml
from fsspec.spec import AbstractFileSystem
from hamilton.function_modifiers import source, value
from munch import Munch, munchify, unmunchify
# from dataclasses import asdict, dataclass, field
from pydantic import BaseModel, ConfigDict, Field

from .helpers.filesystem import get_filesystem


class BaseConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_dict(self) -> dict[str, Any]:
        return unmunchify(self.model_dump())

    def to_yaml(self, path: str, fs: AbstractFileSystem | None = None) -> None:
        try:
            with fs.open(path, "w") as f:
                yaml.dump(self.to_dict(), f, default_flow_style=False)
        except NotImplementedError:
            raise NotImplementedError(
                "The filesystem "
                f"{self.fs.fs.protocol[0] if isinstance(self.fs.fs.protocol, tuple) else self.fs.fs.protocol} "
                "does not support writing files."
            )

    @classmethod
    def from_dict(cls, d: dict[str, Any] | Munch) -> "BaseConfig":
        return cls(**d)

    @classmethod
    def from_yaml(cls, path: str, fs: AbstractFileSystem):
        # if fs is None:
        #    fs = get_filesystem(".", cached=True)
        with fs.open(path) as f:
            return cls.from_dict(yaml.full_load(f))

    def update(self, d: dict[str, Any] | Munch) -> None:
        for k, v in d.items():
            setattr(self, k, v)


class PipelineRunConfig(BaseConfig):
    final_vars: list[str] = Field(default_factory=list)
    inputs: dict | Munch = Field(default_factory=dict)
    executor: str | None = Field(default=None)
    with_tracker: bool = Field(default=False)
    with_opentelemetry: bool = Field(default=False)

    def model_post_init(self, __context):
        if isinstance(self.inputs, dict):
            self.inputs = munchify(self.inputs)


class PipelineScheduleTriggerConfig(BaseConfig):
    type_: str | None = Field(default=None)
    crontab: str | None = Field(default=None)
    year: str | int | None = Field(default=None)
    month: str | int | None = Field(default=None)
    weeks: int | float = Field(default=0)
    week: str | int | None = Field(default=None)
    days: int | float = Field(default=0)
    day: str | int | None = Field(default=None)
    day_of_week: str | int | None = Field(default=None)
    hours: int | float = Field(default=0)
    hour: str | int | None = Field(default=None)
    minutes: int | float = Field(default=0)
    minute: str | int | None = Field(default=None)
    seconds: int | float = Field(default=0)
    second: str | int | None = Field(default=None)
    start_time: dt.datetime | None = Field(default=None)
    end_time: dt.datetime | None = Field(default=None)
    timezone: str | None = Field(default=None)


class PipelineScheduleRunConfig(BaseConfig):
    id_: str | None = Field(default=None)
    executor: str | None = Field(default=None)
    paused: bool = Field(default=False)
    coalesce: str = Field(default="latest")  # other options are "all" and "earliest"
    misfire_grace_time: int | float | dt.timedelta | None = Field(default=None)
    max_jitter: int | float | dt.timedelta | None = Field(default=None)
    max_running_jobs: int | None = Field(default=None)
    conflict_policy: str | None = Field(
        default="do_nothing"
    )  # other options are "replace" and "exception"


class PipelineScheduleConfig(BaseConfig):
    run: PipelineScheduleRunConfig = Field(default_factory=PipelineScheduleRunConfig)
    trigger: PipelineScheduleTriggerConfig = Field(
        default_factory=PipelineScheduleTriggerConfig
    )


class PipelineTrackerConfig(BaseConfig):
    project_id: int | None = Field(default=None)
    version: str | None = Field(default=None)
    dag_name: str | None = Field(default=None)
    tags: dict | Munch = Field(default_factory=dict)

    def model_post_init(self, __context):
        self.tags = munchify(self.tags)


class PipelineConfig(BaseConfig):
    name: str | None = Field(default=None)
    run: PipelineRunConfig = Field(default_factory=PipelineRunConfig)
    schedule: PipelineScheduleConfig = Field(default_factory=PipelineScheduleConfig)
    params: dict | Munch = Field(default_factory=dict)
    tracker: PipelineTrackerConfig = Field(default_factory=PipelineTrackerConfig)
    h_params: dict | Munch = Field(default_factory=dict)

    def model_post_init(self, __context):
        if isinstance(self.params, dict):
            self.h_params = munchify(self.to_h_params(self.params))
            self.params = munchify(self.params)

    def to_yaml(self, path: str, fs: AbstractFileSystem):
        try:
            with fs.open(path, "w") as f:
                d = self.to_dict()
                d.pop("name")
                yaml.dump(d, f, default_flow_style=False)
        except NotImplementedError:
            raise NotImplementedError(
                "The filesystem "
                f"{self.fs.fs.protocol[0] if isinstance(self.fs.fs.protocol, tuple) else self.fs.fs.protocol} "
                "does not support writing files."
            )

    @classmethod
    def from_dict(cls, name: str, d: dict | Munch):
        d.update({"name": name})
        return cls(**d)

    @classmethod
    def from_yaml(cls, name: str, path: str, fs: AbstractFileSystem):
        with fs.open(path) as f:
            return cls.from_dict(name, yaml.full_load(f))

    def update(self, d: dict | Munch):
        for k, v in d.items():
            eval(f"self.{k}.update({v})")
            if k == "params":
                self.params.update(munchify(v))
                self.h_params = munchify(self.to_h_params(self.params))
                # self.params = munchify(self.params)
        if "params" in d:
            self.h_params = munchify(self.to_h_params(self.params))
            self.params = munchify(self.params)

    @staticmethod
    def to_h_params(d: dict) -> dict:
        """Coverts a dictionary of function arguments to Hamilton function parameters"""

        def transform_recursive(val, original_dict):
            # If it's a dictionary, recursively transform its values
            if isinstance(val, dict):
                return {
                    k: transform_recursive(v, original_dict) for k, v in val.items()
                }
            # If it's a string and matches a key in the original dictionary
            elif isinstance(val, str) and val in original_dict:
                return source(val)
            # For all other values
            else:
                return value(val)

        # Step 1: Replace each value with a dictionary containing key and value
        result = {k: {k: d[k]} for k in d}

        # Step 2 & 3: Transform all values recursively
        return {k: transform_recursive(v, d) for k, v in result.items()}


class ProjectWorkerConfig(BaseConfig):
    data_store: dict | Munch = Field(default_factory=dict)
    event_broker: dict | Munch = Field(default_factory=dict)
    cleanup_interval: int | float | dt.timedelta = Field(default=900)  # int in secods
    max_concurrent_jobs: int = Field(default=100)

    def model_post_init(self, __context):
        if isinstance(self.data_store, dict):
            self.data_store = munchify(self.data_store)
        if isinstance(self.event_broker, dict):
            self.event_broker = munchify(self.event_broker)


class ProjectTrackerConfig(BaseConfig):
    username: str | None = Field(default=None)
    api_url: str = "http://localhost:8241"
    ui_url: str = "http://localhost:8242"
    api_key: str | None = Field(default=None)


class ProjectOpenTelemetryConfig(BaseConfig):
    host: str = Field(default="localhost")
    port: int = Field(default=6831)


class ProjectConfig(BaseConfig):
    name: str | None = Field(default=None)
    worker: ProjectWorkerConfig = Field(default_factory=ProjectWorkerConfig)
    tracker: ProjectTrackerConfig = Field(default_factory=ProjectTrackerConfig)
    open_telemetry: ProjectOpenTelemetryConfig = Field(
        default_factory=ProjectOpenTelemetryConfig
    )


class Config(BaseConfig):
    pipeline: PipelineConfig = Field(default_factory=PipelineConfig)
    project: ProjectConfig = Field(default_factory=ProjectConfig)
    fs: AbstractFileSystem | None = Field(default=None)
    base_dir: str | pathlib.Path | None = Field(default=None)
    storage_options: dict | Munch = Field(default_factory=Munch)

    @classmethod
    def load(
        cls,
        base_dir: str = "",
        name: str | None = None,
        pipeline_name: str | None = None,
        fs: AbstractFileSystem | None = None,
        storage_options: dict | Munch = Munch(),
    ):
        if fs is None:
            fs = get_filesystem(base_dir, cached=True, dirfs=True, **storage_options)
        if fs.exists("conf/project.yml"):
            project = ProjectConfig.from_yaml(path="conf/project.yml", fs=fs)
        else:
            project = ProjectConfig(name=name)

        if pipeline_name is not None:
            if fs.exists(f"conf/pipelines/{pipeline_name}.yml"):
                pipeline = PipelineConfig.from_yaml(
                    name=pipeline_name,
                    path=f"conf/pipelines/{pipeline_name}.yml",
                    fs=fs,
                )
            else:
                pipeline = PipelineConfig(name=pipeline_name)
        else:
            pipeline = PipelineConfig(name=pipeline_name)

        return cls(
            base_dir=base_dir,
            pipeline=pipeline,
            project=project,
            fs=fs,
            storage_options=storage_options,
        )

    def save(self):
        if not self.fs.exists("conf"):
            self.fs.makedirs("conf")

        if self.pipeline.name is not None:
            self.pipeline.to_yaml(f"conf/pipelines/{self.pipeline.name}.yml", self.fs)
        self.project.to_yaml("conf/project.yml", self.fs)


def load(
    base_dir: str,
    name: str | None = None,
    pipeline_name: str | None = None,
    storage_options: dict | Munch = Munch(),
    fs: AbstractFileSystem | None = None,
):
    return Config.load(
        name=name,
        pipeline_name=pipeline_name,
        base_dir=base_dir,
        storage_options=storage_options,
        fs=fs,
    )


def save(config: Config):
    config.save()



import datetime as dt
import importlib
import importlib.util
import os
import sys
from typing import Any, Callable
from uuid import UUID

import rich
from fsspec.spec import AbstractFileSystem
from hamilton import driver
from hamilton.plugins import h_opentelemetry
from hamilton_sdk.adapters import HamiltonTracker
from loguru import logger
from rich.columns import Columns
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.tree import Tree

from .cfg import (Config, PipelineConfig, PipelineRunConfig,
                  PipelineScheduleConfig, PipelineTrackerConfig)
from .helpers.filesystem import get_filesystem
from .helpers.open_telemetry import init_tracer
# from munch import unmunchify
from .helpers.templates import PIPELINE_PY_TEMPLATE

if importlib.util.find_spec("apscheduler"):
    from .scheduler import SchedulerManager
else:
    SchedulerManager = None
if importlib.util.find_spec("paho"):
    from .helpers.mqtt import MQTTClient
else:
    MQTTClient = None


from .helpers.executor import get_executor
from .helpers.trigger import get_trigger  # , ALL_TRIGGER_KWARGS


class PipelineManager:
    def __init__(
        self,
        base_dir: str | None = None,
        storage_options: dict = {},
        fs: AbstractFileSystem | None = None,
    ):
        """
        Initializes the Pipeline object.

        Args:
            base_dir (str | None): The flowerpower base path. Defaults to None.
            storage_options (dict, optional): The fsspec storage options. Defaults to {}.
            fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.

        Returns:
            None
        """
        self._base_dir = base_dir or ""
        self._storage_options = storage_options
        if fs is None:
            fs = get_filesystem(self._base_dir, **self._storage_options)
        self._fs = fs

        self._conf_dir = "conf"
        self._pipeline_dir = "pipelines"

        self._sync_fs()
        self.load_config()

    def _get_schedules(self):
        with SchedulerManager(
            # name=f"{self.cfg.project.name}.{self.name}",
            fs=self._fs,
            role="scheduler",
        ) as sm:
            return sm.get_schedules()

    def _sync_fs(self):
        """
        Sync the filesystem.

        Returns:
            None
        """
        if self._fs.is_cache_fs:
            self._fs.sync()

        modules_path = os.path.join(self._fs.path, self._pipeline_dir)
        if modules_path not in sys.path:
            sys.path.append(modules_path)

    def load_module(self, name: str):
        """
        Load a module dynamically.

        Args:
            name (str): The name of the module to load.

        Returns:
            None
        """
        sys.path.append(os.path.join(self._fs.path, self._pipeline_dir))

        if not hasattr(self, "_module"):
            self._module = importlib.import_module(name)
        else:
            self._module = importlib.reload(self._module)

    def load_config(self, name: str | None = None):
        """
        Load the configuration file.

        This method loads the configuration file specified by the `_conf_dir` attribute and
        assigns it to the `cfg` attribute.

        Args:
            name (str | None, optional): The name of the pipeline. Defaults to None.

        Returns:
            None
        """
        self.cfg = Config.load(base_dir=self._base_dir, pipeline_name=name, fs=self._fs)

    def _get_driver(
        self,
        name: str,
        executor: str | None = None,
        with_tracker: bool = False,
        with_opentelemetry: bool = False,
        reload: bool = False,
        **kwargs,
    ) -> tuple[driver.Driver, Callable | None]:
        """
        Get the driver and shutdown function for a given pipeline.

        Args:
            name (str): The name of the pipeline.
            executor (str | None, optional): The executor to use. Defaults to None.
            with_tracker (bool, optional): Whether to use the tracker. Defaults to False.
            with_opentelemetry (bool, optional): Whether to use OpenTelemetry. Defaults to False.
            reload (bool, optional): Whether to reload the module. Defaults to False.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            max_tasks (int, optional): The maximum number of tasks. Defaults to 20.
            num_cpus (int, optional): The number of CPUs. Defaults to 4.
            project_id (str, optional): The project ID for the tracker. Defaults to None.
            username (str, optional): The username for the tracker. Defaults to None.
            dag_name (str, optional): The DAG name for the tracker. Defaults to None.
            tags (str, optional): The tags for the tracker. Defaults to None.
            api_url (str, optional): The API URL for the tracker. Defaults to None.
            ui_url (str, optional): The UI URL for the tracker. Defaults to None.

        Returns:
            tuple[driver.Driver, Callable | None]: A tuple containing the driver and shutdown function.
        """
        if not self.cfg.pipeline.name == name:
            self.load_config(name=name)

        max_tasks = kwargs.pop("max_tasks", 20)
        num_cpus = kwargs.pop("num_cpus", 4)
        executor_, shutdown = get_executor(
            executor or "local", max_tasks=max_tasks, num_cpus=num_cpus
        )
        if reload or not hasattr(self, "_module"):
            self.load_module(name=name)

        adapters = []
        if with_tracker:
            tracker_cfg = {
                **self.cfg.pipeline.tracker.to_dict(),
                **self.cfg.project.tracker.to_dict(),
            }
            tracker_kwargs = {
                key: kwargs.pop(key, None) or tracker_cfg.get(key, None)
                for key in tracker_cfg
            }
            tracker_kwargs["hamilton_api_url"] = tracker_kwargs.pop("api_url", None)
            tracker_kwargs["hamilton_ui_url"] = tracker_kwargs.pop("ui_url", None)

            if tracker_kwargs.get("project_id", None) is None:
                raise ValueError(
                    "Please provide a project_id if you want to use the tracker"
                )

            tracker = HamiltonTracker(**tracker_kwargs)
            adapters.append(tracker)

        if with_opentelemetry:
            trace = init_tracer(
                host=kwargs.pop("host", "localhost"),
                port=kwargs.pop("port", 6831),
                name=f"{self.cfg.project.name}.{name}",
            )
            tracer = trace.get_tracer(__name__)
            adapters.append(h_opentelemetry.OpenTelemetryTracer(tracer=tracer))
        if len(adapters):
            print("adapters len:", len(adapters))

            dr = (
                driver.Builder()
                .with_modules(self._module)
                .enable_dynamic_execution(allow_experimental_mode=True)
                .with_adapters(adapters[0])
                .with_remote_executor(executor_)
                .build()
            )
        else:
            dr = (
                driver.Builder()
                .with_modules(self._module)
                .enable_dynamic_execution(allow_experimental_mode=True)
                .with_remote_executor(executor_)
                .build()
            )

        return dr, shutdown

    def _run(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        """
        Run the pipeline with the given parameters.

        Args:
            name (str): The name of the pipeline.
            executor (str | None, optional): The executor to use. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker. Defaults to None.
            with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry. Defaults to None.
            reload (bool, optional): Whether to reload the pipeline. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The result of executing the pipeline.
        """
        if not self.cfg.pipeline.name == name:
            self.load_config(name=name)

        logger.info(
            f"Starting pipeline {self.cfg.project.name}.{name}"
        )  # in environment {environment}")

        run_params = self.cfg.pipeline.run

        final_vars = final_vars or run_params.final_vars
        inputs = {
            **(run_params.inputs or {}),
            **(inputs or {}),
        }  # <-- inputs override and adds to run_params

        kwargs.update(
            {
                arg: eval(arg) or getattr(run_params, arg)
                for arg in ["executor", "with_tracker", "with_opentelemetry"]
            }
        )

        dr, shutdown = self._get_driver(
            name=name,
            **kwargs,
        )

        res = dr.execute(final_vars=final_vars, inputs=inputs)

        logger.success(f"Finished pipeline {self.cfg.project.name}.{name}")

        if shutdown is not None:
            shutdown()

        return res

    def run(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool = False,
        with_opentelemetry: bool = False,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        """
        Run the pipeline.

        Args:
            name (str): The name of the pipeline.
            executor (str | None, optional): The executor to use. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            with_tracker (bool, optional): Whether to use a tracker. Defaults to False.
            with_opentelemetry (bool, optional): Whether to use OpenTelemetry. Defaults to False.
            reload (bool, optional): Whether to reload the pipeline. Defaults to False.
            **kwargs: Additional keyword arguments.
        Returns:
            Any: The result of running the pipeline.
        """
        return self._run(
            name=name,
            inputs=inputs,
            final_vars=final_vars,
            executor=executor,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
            **kwargs,
        )

    def run_job(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        """
        Add a job to run the pipeline with the given parameters to the worker.
        Executes the job immediatly and returns the result of the execution.

        Args:
            name (str): The name of the job.
            executor (str | None, optional): The executor to use for the job. Defaults to None.
            inputs (dict | None, optional): The inputs for the job. Defaults to None.
            final_vars (list | None, optional): The final variables for the job. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
            with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the job. Defaults to None.
            reload (bool, optional): Whether to reload the job. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The result of the job execution.
        """

        with SchedulerManager(
            name=f"{self.cfg.project.name}.{name}",
            fs=self._fs,
            role="scheduler",
        ) as sm:
            # if not any([task.id == "run-pipeline" for task in sm.get_tasks()]):
            #    sm.configure_task(func_or_task_id="run-pipeline", func=self._run)
            kwargs.update(
                {
                    arg: eval(arg)
                    for arg in [
                        "name",
                        "inputs",
                        "final_vars",
                        "executor",
                        "with_tracker",
                        "with_opentelemetry",
                        "reload",
                    ]
                }
            )
            return sm.run_job(
                self._run,
                # args=(
                #     name,
                #     inputs,
                #     final_vars,
                #     executor,
                #     with_tracker,
                #     with_opentelemetry,
                #     reload,
                # ),
                kwargs=kwargs,
                job_executor=(
                    executor
                    if executor in ["async", "threadpool", "processpool"]
                    else "threadpool"
                ),
            )

    def _add_job(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        **kwargs,
    ) -> UUID:
        """
        Add a job to run the pipeline with the given parameters to the worker data store.
        Executes the job immediatly and returns the job id (UUID). The job result will be stored in the data store
        for the given `result_expiration_time` and can be fetched using the job id (UUID).

        Args:
            name (str): The name of the job.
            executor (str | None, optional): The executor for the job. Defaults to None.
            inputs (dict | None, optional): The inputs for the job. Defaults to None.
            final_vars (list | None, optional): The final variables for the job. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
            with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the job. Defaults to None.
            reload (bool, optional): Whether to reload the job. Defaults to False.
            result_expiration_time (float | dt.timedelta | None, optional): The result expiration time for the job.
                Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            UUID: The UUID of the added job.
        """
        with SchedulerManager(
            name=f"{self.cfg.project.name}.{name}",
            fs=self._fs,
            role="scheduler",
        ) as sm:
            kwargs.update(
                {
                    arg: eval(arg)
                    for arg in [
                        "name",
                        "inputs",
                        "final_vars",
                        "executor",
                        "with_tracker",
                        "with_opentelemetry",
                        "reload",
                    ]
                }
            )
            id_ = sm.add_job(
                self._run,
                # args=(
                #     name,
                #     inputs,
                #     final_vars,
                #     executor,
                #     with_tracker,
                #     with_opentelemetry,
                #     reload,
                # ),
                kwargs=kwargs,
                job_executor=(
                    executor
                    if executor in ["async", "threadpool", "processpool"]
                    else "threadpool"
                ),
                result_expiration_time=result_expiration_time,
            )
            rich.print(
                f"✅ Successfully added job for "
                f"[blue]{self.cfg.project.name}.{name}[/blue] with ID [green]{id_}[/green]"
            )
            return id_

    def add_job(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        **kwargs,
    ):
        """
        Add a job to run the pipeline with the given parameters to the worker data store.
        Executes the job immediatly and returns the job id (UUID). The job result will be stored in the data store
        for the given `result_expiration_time` and can be fetched using the job id (UUID).

        Args:
            name (str): The name of the job.
            executor (str | None, optional): The executor for the job. Defaults to None.
            inputs (dict | None, optional): The inputs for the job. Defaults to None.
            final_vars (list | None, optional): The final variables for the job. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
            with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the job. Defaults to None.
            reload (bool, optional): Whether to reload the job. Defaults to False.
            result_expiration_time (float | dt.timedelta | None, optional): The result expiration time for the job.
                Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            UUID: The UUID of the added job.
        """
        return self._add_job(
            name=name,
            inputs=inputs,
            final_vars=final_vars,
            executor=executor,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
            result_expiration_time=result_expiration_time,
            **kwargs,
        )

    def schedule(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        trigger_type: str | None = None,
        id_: str | None = None,
        paused: bool = False,
        coalesce: str = "latest",
        misfire_grace_time: float | dt.timedelta | None = None,
        max_jitter: float | dt.timedelta | None = None,
        max_running_jobs: int | None = None,
        conflict_policy: str = "do_nothing",
        # job_result_expiration_time: float | dt.timedelta = 0,
        overwrite: bool = False,
        **kwargs,
    ) -> str:
        """
        Schedule a pipeline for execution.

        Args:
            name (str): The name of the pipeline.
            executor (str | None, optional): The executor to use for running the pipeline. Defaults to None.
            trigger_type (str | None, optional): The type of trigger for the pipeline. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            with_tracker (bool | None, optional): Whether to include a tracker for the pipeline. Defaults to None.
            with_opentelemetry (bool | None, optional): Whether to include OpenTelemetry for the pipeline.
                Defaults to None.
            id_ (str | None, optional): The ID of the scheduled pipeline. Defaults to None.
            paused (bool, optional): Whether the pipeline should be initially paused. Defaults to False.
            coalesce (str, optional): The coalesce strategy for the pipeline. Defaults to "latest".
            misfire_grace_time (float | dt.timedelta | None, optional): The grace time for misfired jobs.
                Defaults to None.
            max_jitter (float | dt.timedelta | None, optional): The maximum number of seconds to randomly add to the
                scheduled. Defaults to None.
            max_running_jobs (int | None, optional): The maximum number of running jobs for the pipeline.
                Defaults to None.
            conflict_policy (str, optional): The conflict policy for the pipeline. Defaults to "do_nothing".
            job_result_expiration_time (float | dt.timedelta | None, optional): The result expiration time for the job.
                Defaults to None.
            overwrite (bool, optional): Whether to overwrite an existing schedule with the same name. Defaults to False.
            **kwargs: Additional keyword arguments for the trigger.

        Returns:
            str: The ID of the scheduled pipeline.

        Raises:
            ValueError: If APScheduler4 is not installed.

        """
        if SchedulerManager is None:
            raise ValueError("APScheduler4 not installed. Please install it first.")

        # if "pipeline" in self.cfg.scheduler:
        schedule_cfg = self.cfg.pipeline.schedule  # .copy()
        run_cfg = self.cfg.pipeline.run

        kwargs.update(
            {arg: eval(arg) or getattr(run_cfg, arg) for arg in run_cfg.to_dict()}
        )
        trigger_kwargs = {
            key: kwargs.pop(key, None) or getattr(schedule_cfg.trigger, key)
            for key in schedule_cfg.trigger.to_dict()
        }
        trigger_type = trigger_type or schedule_cfg.trigger.type_
        trigger_kwargs.pop("type_", None)

        schedule_kwargs = {
            arg: eval(arg) or getattr(schedule_cfg.run, arg)
            for arg in schedule_cfg.run.to_dict()
        }
        executor = executor or schedule_cfg.run.executor
        # id_ = id_ or schedule_cfg.run.id_

        def _get_id() -> str:
            if id_:
                return id_

            if overwrite:
                return f"{name}-1"

            ids = [schedule.id for schedule in self._get_schedules()]
            if any([name in id_ for id_ in ids]):
                id_num = sorted([id_ for id_ in ids if name in id_])[-1].split("-")[-1]
                return f"{name}-{int(id_num) + 1}"
            return f"{name}-1"

        id_ = _get_id()

        schedule_kwargs.pop("executor", None)
        schedule_kwargs.pop("id_", None)

        with SchedulerManager(
            name=f"{self.cfg.project.name}.{name}",
            fs=self._fs,
            role="scheduler",
        ) as sm:
            trigger = get_trigger(type_=trigger_type, **trigger_kwargs)

            id_ = sm.add_schedule(
                func_or_task_id=self._run,
                trigger=trigger,
                id=id_,
                args=(name,),  # inputs, final_vars, executor, with_tracker),
                kwargs=kwargs,
                job_executor=(
                    executor
                    if executor in ["async", "threadpool", "processpool"]
                    else "threadpool"
                ),
                **schedule_kwargs,
            )
            rich.print(
                f"✅ Successfully added schedule for "
                f"[blue]{self.cfg.project.name}.{name}[/blue] with ID [green]{id_}[/green]"
            )
            return id_

    def add(
        self,
        name: str | None = None,
        overwrite: bool = False,
        pipeline_file: str | None = None,
        pipeline_config: str | dict | PipelineConfig = {},
    ):
        """
        Adds a pipeline with the given name.

        Args:
            name (str): The name of the pipeline.
            overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
            pipeline_file (str | None, optional): The path to the pipeline file or the pipeline file content.
                Defaults to None.
            pipeline_config (str | dict | PipelineConfig, optional): The configuration for the pipeline or the pipeline config content.
                Defaults to {}.
        """

        if not self._fs.exists(self._conf_dir):
            raise ValueError(
                f"Configuration path {self._conf_dir} does not exist. Please run flowerpower init first."
            )
        if not self._fs.exists(self._pipeline_dir):
            raise ValueError(
                f"Pipeline path {self._pipeline_dir} does not exist. Please run flowerpower init first."
            )

        if self._fs.exists(f"{self._pipeline_dir}/{name}.py") and not overwrite:
            raise ValueError(
                f"Pipeline {self.cfg.project.name}.{name} already exists. Use `overwrite=True` to overwrite."
            )
        if name is None:
            if not pipeline_file:
                raise ValueError("Please provide a name for the pipeline.")
            name = os.path.splitext(os.path.basename(pipeline_file))[0]
        try:
            if pipeline_file:
                if ".py" not in pipeline_file:
                    self._fs.put_file(pipeline_file, f"{self._pipeline_dir}/{name}.py")
                else:
                    self._fs.write_text(
                        f"{self._pipeline_dir}/{name}.py", pipeline_file
                    )
            else:
                with self._fs.open(f"{self._pipeline_dir}/{name}.py", "w") as f:
                    f.write(
                        PIPELINE_PY_TEMPLATE.format(
                            name=name,
                            date=dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        )
                    )
        except NotImplementedError:
            raise NotImplementedError(
                "The filesystem "
                f"{self.fs.fs.protocol[0] if isinstance(self.fs.fs.protocol, tuple) else self.fs.fs.protocol} "
                "does not support writing files."
            )

        if pipeline_config:
            if isinstance(pipeline_config, str):
                if ".yml" in pipeline_config or ".yaml" in pipeline_config:
                    self._fs.put_file(
                        pipeline_config, f"{self._conf_dir}/pipelines/{name}.yml"
                    )
                else:
                    self._fs.write_text(
                        f"{self._conf_dir}/pipelines/{name}.yml", pipeline_config
                    )
            else:
                self.cfg.pipeline = (
                    pipeline_config
                    if isinstance(pipeline_config, PipelineConfig)
                    else PipelineConfig.from_dict(name, pipeline_config)
                )
                if not self.cfg.pipeline.name:
                    self.cfg.pipeline.name = name
        else:
            self.cfg.pipeline = PipelineConfig(name=name)

        self.cfg.save()

        rich.print(
            f"🔧 Created new pipeline [bold blue]{self.cfg.project.name}.{name}[/bold blue]"
        )

    def new(
        self,
        name: str,
        overwrite: bool = False,
    ):
        """
        Adds a new pipeline with the given name.

        Args:
            name (str): The name of the pipeline.
            overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        """
        self.add(name, overwrite)

    def delete(self, name: str, cfg: bool = True, remove_module_file: bool = False):
        """
        Delete a pipeline.

        Args:
            name (str): The name of the pipeline to delete.
            cfg (bool, optional): Whether to delete the pipeline configuration. Defaults to True.
            remove_module_file (bool, optional): Whether to delete the pipeline module file. Defaults to False.
        """

        if cfg:
            if self._fs.exists(f"{self._conf_dir}/pipelines/{name}.yml"):
                self._fs.rm(f"{self._conf_dir}/pipelines/{name}.yml")
                rich.print(f"🗑️ Deleted pipeline config for {name}")

        if remove_module_file:
            if self._fs.exists(f"{self._pipeline_dir}/{name}.py"):
                self._fs.rm(f"{self._pipeline_dir}/{name}.py")
                rich.print(
                    f"🗑️ Deleted pipeline config for {self.cfg.project.name}.{name}"
                )

    def show_dag(
        self,
        name: str,
        format: str = "png",
        show: bool = False,
        reload: bool = False,
        save: bool = False,
        **kwargs,
    ):
        """
        Display the graph of functions for a given name.

        Args:
            name (str): The name of the graph.
            format (str, optional): The format of the graph file. Defaults to "png".
            show (bool, optional): Whether to open the graph file after generating it. Defaults to False.
            reload (bool, optional): Whether to reload the graph data. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            graph: The generated graph object.
        """
        self._fs.makedirs("graphs", exist_ok=True)
        dr, _ = self._get_driver(
            name=name, executor=None, with_tracker=False, reload=reload
        )
        if save:
            path = f"graphs/{name}.{format}"
        else:
            path = None
        graph = dr.display_all_functions(path, **kwargs)

        if show:
            graph.view()
        return graph

    def _get_files(self) -> list[str]:
        """
        Get the pipeline files.

        Returns:
            list[str]: A list of pipeline files.
        """
        return [f for f in self._fs.ls(self._pipeline_dir) if f.endswith(".py")]

    def _get_names(self) -> list[str]:
        """
        Get the pipeline names.

        Returns:
            list[str]: A list of pipeline names.
        """
        return [os.path.splitext(os.path.basename(f))[0] for f in self._get_files()]

    def get_summary(
        self, name: str | None = None, config: bool = True, module: bool = True
    ) -> dict:
        """
        Get a summary of the pipelines.

        Args:
            name (str | None, optional): The name of the pipeline. Defaults to None.
            config (bool, optional): Whether to show the configuration. Defaults to True.
            module (bool, optional): Whether to show the module. Defaults to True.
        Returns:
            dict: A dictionary containing the pipeline summary.
        """
        if name:
            pipeline_names = [name]
        else:
            pipeline_names = self._get_names()

        pipeline_summary = {}
        for name in pipeline_names:
            self.load_config(name)
            if config:
                pipeline_summary[name] = {"config": self.cfg.pipeline.to_dict()}
            if module:
                pipeline_summary[name].update(
                    {
                        "module": self._fs.cat(
                            f"{self._pipeline_dir}/{name}.py"
                        ).decode(),
                    }
                )
        return pipeline_summary

    def show_summary(
        self, name: str | None = None, config: bool = True, module: bool = True
    ) -> dict:
        """
        Show a summary of the pipelines.

        Args:
            name (str | None, optional): The name of the pipeline. Defaults to None.
            config (bool, optional): Whether to show the configuration. Defaults to True.
            module (bool, optional): Whether to show the module. Defaults to True.
        """

        pipeline_summary = self.get_summary(name=name, config=config, module=module)

        def add_dict_to_tree(tree, dict_data, style="green"):
            for key, value in dict_data.items():
                if isinstance(value, dict):
                    branch = tree.add(f"[cyan]{key}:", style="bold cyan")
                    add_dict_to_tree(branch, value, style)
                else:
                    tree.add(f"[cyan]{key}:[/] [green]{value}[/]")

        console = Console()
        for pipeline, info in pipeline_summary.items():
            # Create tree for config
            config_tree = Tree("📋 Configuration", style="bold magenta")
            add_dict_to_tree(config_tree, info["config"])

            # Create syntax-highlighted code view
            code_view = Syntax(
                info["module"],
                "python",
                theme="default",
                line_numbers=False,
                word_wrap=True,
                code_width=80,
                padding=2,
            )

            # console.print(
            #     Columns(
            #         [
            #             Panel(
            #                 config_tree,
            #                 title=f"🔄 Pipeline: {pipeline}",
            #                 subtitle="Configuration",
            #                 border_style="blue",
            #                 padding=(2, 2),
            #             ),
            #             Panel(
            #                 code_view,
            #                 title=f"🔄 Pipeline: {pipeline}",
            #                 subtitle="Module",
            #                 border_style="blue",
            #                 padding=(2, 2),
            #             ),
            #         ],
            #         # column_first=True,
            #         equal=True,
            #         expand=True,
            #     )
            # )
            # console.print("\n")
            if config:
                # console.print(f"🔄 Pipeline: {pipeline}", style="bold blue")
                console.print(
                    Panel(
                        config_tree,
                        title=f"🔄 Pipeline: {pipeline}",
                        subtitle="Configuration",
                        border_style="blue",
                        padding=(2, 2),
                    )
                )
                console.print("\n")

            if module:
                # console.print(f"🔄 Pipeline: {pipeline}", style="bold blue")
                console.print(
                    Panel(
                        code_view,
                        title=f"🔄 Pipeline: {pipeline}",
                        subtitle="Module",
                        border_style="blue",
                        padding=(2, 2),
                    )
                )
                console.print("\n")

    @property
    def summary(self) -> dict:
        """
        Get a summary of the pipelines.

        Returns:
            dict: A dictionary containing the pipeline summary.
        """
        return self.get_summary()

    def all_pipelines(self, show: bool = True) -> None:
        """
        Print all available pipelines in a formatted table.

        Args:
            show (bool, optional): Whether to print the table. Defaults to True.
        Returns:
            None
        """

        pipeline_files = [
            f for f in self._fs.ls(self._pipeline_dir) if f.endswith(".py")
        ]
        pipeline_names = [os.path.splitext(f)[0] for f in pipeline_files]

        if not pipeline_files:
            rich.print("[yellow]No pipelines found[/yellow]")
            return

        pipeline_info = []

        for path, name in zip(pipeline_files, pipeline_names):
            # path = os.path.join( f)
            try:
                mod_time = self._fs.modified(path).strftime("%Y-%m-%d %H:%M:%S")
            except NotImplementedError:
                mod_time = "N/A"
            size = f"{self._fs.size(path) / 1024:.1f} KB"
            pipeline_info.append(
                {"name": name, "path": path, "mod_time": mod_time, "size": size}
            )

        if show:
            table = Table(title="Available Pipelines")
            table.add_column("Pipeline Name", style="blue")
            table.add_column("Path", style="magenta")
            table.add_column("Last Modified", style="green")
            table.add_column("Size", style="cyan")

            for info in pipeline_info:
                table.add_row(
                    info["name"], info["path"], info["mod_time"], info["size"]
                )
            rich.print(table)
        else:
            return pipeline_info

    def start_mqtt_listener(
        self,
        name: str,
        topic: str | None = None,
        host: str = "localhost",
        port: int = 1883,
        user: str | None = None,
        pw: str | None = None,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        as_job: bool = False,
        background: bool = False,
        **kwargs,
    ):
        """
        Run a pipeline when a message is received on a given topic.

        Args:
            name (str): The name of the pipeline.
            topic (str | None, optional): The topic to subscribe to. Defaults to None.
            host (str, optional): The host of the MQTT broker. Defaults to "localhost".
            port (int, optional): The port of the MQTT broker. Defaults to 1883.
            user (str | None, optional): The username for the MQTT broker. Defaults to None.
            pw (str | None, optional): The password for the MQTT broker. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            executor (str | None, optional): The executor to use for the pipeline. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker for the pipeline. Defaults to None.
            with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the pipeline. Defaults to None.
            reload (bool, optional): Whether to reload the pipeline. Defaults to False.
            result_expiration_time (float | dt.timedelta, optional): The result expiration time for the job.
                Defaults to 0.
            as_job (bool, optional): Whether to run the pipeline as a job. Defaults to False.
            background (bool, optional): Whether to run the pipeline in the background. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            None
        """
        if inputs is None:
            inputs = {}

        def on_message(client, userdata, msg):
            logger.info("Message arrived")

            inputs["payload"] = msg.payload

            try:
                if as_job:
                    self._add_job(
                        name=name,
                        inputs=inputs,
                        final_vars=final_vars,
                        executor=executor,
                        with_tracker=with_tracker,
                        with_opentelemetry=with_opentelemetry,
                        reload=reload,
                        result_expiration_time=result_expiration_time,
                        **kwargs,
                    )
                else:
                    self._run(
                        name=name,
                        inputs=inputs,
                        final_vars=final_vars,
                        executor=executor,
                        with_tracker=with_tracker,
                        with_opentelemetry=with_opentelemetry,
                        reload=reload,
                        result_expiration_time=result_expiration_time,
                        **kwargs,
                    )
                logger.success("Message processed successfully")
                return
            except Exception as e:
                _ = e
                logger.exception(e)

            logger.warning("processing failed")

        if not hasattr(self, "mqtt_client"):
            self.mqtt_client = {}
        try:
            self.mqtt[name] = MQTTClient.from_event_broker(base_dir=self._base_dir)
        except ValueError as e:
            logger.exception(e)
        else:
            self.mqtt[name] = MQTTClient(user=user, pw=pw, host=host, port=port)
        if topic is None:
            topic = name
        self.mqtt[name].connect()
        self.mqtt[name].start_listener(on_message, topic, background=background)

    def stop_mqtt_listener(self, name: str):
        """
        Stop the MQTT listener.

        Returns:
            None
        """
        self.mqtt[name].stop_listener()
        self.mqtt[name].disconnect()


class Pipeline(PipelineManager):
    def __init__(
        self,
        name: str,
        base_dir: str | None = None,
        storage_options: dict = {},
        fs: AbstractFileSystem | None = None,
    ):
        """
        Initializes the Pipeline object.

        Args:
            name (str): The name of the pipeline.
            base_dir (str | None): The flowerpower base path. Defaults to None.
            storage_options (dict, optional): The fsspec storage options. Defaults to {}.
            fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.

        Returns:
            None
        """
        super().__init__(base_dir=base_dir, storage_options=storage_options, fs=fs)
        self.name = name
        self.load_module()
        self.load_config(name)

    def run(
        self,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool = False,
        with_opentelemetry: bool = False,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        name = self.name
        return super().run(
            name=name,
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
            **kwargs,
        )

    def run_job(
        self,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        name = self.name
        return super().run_job(
            name=name,
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
            **kwargs,
        )

    def add_job(
        self,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        **kwargs,
    ) -> UUID:
        name = self.name
        return super().add_job(
            name=name,
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
            result_expiration_time=result_expiration_time,
            **kwargs,
        )

    def schedule(
        self,
        trigger_type: str = "cron",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool = False,
        with_opentelemetry: bool = False,
        paused: bool = False,
        coalesce: str = "latest",
        misfire_grace_time: float | dt.timedelta | None = None,
        max_jitter: float | dt.timedelta | None = None,
        max_running_jobs: int | None = None,
        conflict_policy: str = "do_nothing",
        # job_result_expiration_time: float | dt.timedelta = 0,
        **kwargs,
    ):
        name = self.name
        return super().schedule(
            name=name,
            executor=executor,
            trigger_type=trigger_type,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            paused=paused,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            max_jitter=max_jitter,
            max_running_jobs=max_running_jobs,
            # job_result_expiration_time=job_result_expiration_time,
            conflict_policy=conflict_policy,
            **kwargs,
        )

    def show_dag(
        self,
        format: str = "png",
        show: bool = False,
        save: bool = False,
        reload: bool = False,
    ):
        return super().show_dag(
            self.name, format=format, show=show, save=save, reload=reload
        )

    def delete(self, cfg: bool = True, remove_module_file: bool = False):
        return super().delete(self.name, cfg, remove_module_file)

    def load_module(self):
        super().load_module(self.name)

    def start_mqtt_listener(
        self,
        topic: str | None = None,
        host: str = "localhost",
        port: int = 1883,
        user: str | None = None,
        pw: str | None = None,
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        with_opentelemetry: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        background: bool = False,
        **kwargs,
    ):
        name = self.name
        return super().start_mqtt_listener(
            name=name,
            topic=topic,
            host=host,
            port=port,
            user=user,
            pw=pw,
            inputs=inputs,
            final_vars=final_vars,
            executor=executor,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
            result_expiration_time=result_expiration_time,
            background=background,
            **kwargs,
        )

    def stop_mqtt_listener(self):
        return super().stop_mqtt_listener(self.name)

    def get_summary(self, config: bool = True, module: bool = True):
        return super().get_summary(self.name, config=config, module=module)[self.name]

    def show_summary(self, config: bool = True, module: bool = True):
        return super().show_summary(self.name, config=config, module=module)

    @property
    def summary(self):
        return self.get_summary()


def add_pipeline(
    name: str,
    overwrite: bool = False,
    pipeline_file: str | None = None,
    pipeline_config: str | dict | PipelineConfig = {},
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    """
    Add a new pipeline.

    Args:
        name (str): The name of the pipeline.
        overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        pipeline_file (str | None, optional): The path to the pipeline file. Defaults to None.
        pipeline_config (str | dict | PipelineConfig, optional): The pipeline configuration. Defaults to {}.
        base_dir (str | None, optional): The base directory of the pipeline. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
    """
    pm = PipelineManager(base_dir=base_dir, storage_options=storage_options, fs=fs)
    pm.add(
        name=name,
        overwrite=overwrite,
        pipeline_file=pipeline_file,
        pipeline_config=pipeline_config,
    )


def add_pipeline_job(
    name: str,
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool | None = None,
    with_opentelemetry: bool | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    reload: bool = False,
    result_expiration_time: float | dt.timedelta = 0,
    **kwargs,
) -> UUID:
    """
    Add a job to run the pipeline with the given parameters to the worker data store.
    Executes the job immediatly and returns the job id (UUID). The job result will be stored in the data store for the
    given `result_expiration_time` and can be fetched using the job id (UUID).

    Args:
        name (str): The name of the job.
        executor (str | None, optional): The executor to use for the job. Defaults to None.
        inputs (dict | None, optional): The inputs for the job. Defaults to None.
        final_vars (list | None, optional): The final variables for the job. Defaults to None.
        with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
        with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the job. Defaults to None.
        base_dir (str | None, optional): The base path for the job. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
        reload (bool, optional): Whether to reload the job. Defaults to False.
        result_expiration_time (float | dt.timedelta | None, optional): The expiration time for the job result.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        UUID: The UUID of the added job.
    """
    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    return p.add_job(
        executor=executor,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        reload=reload,
        result_expiration_time=result_expiration_time,
        **kwargs,
    )


def all_pipelines(
    base_dir: str | None = None,
    show: bool = True,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    """
    Print all available pipelines in a formatted table.

    Args:
        base_dir (str | None, optional): The base path of the pipelines. Defaults to None.
        show (bool, optional): Whether to print the table. Defaults to True.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
    """
    pm = PipelineManager(base_dir=base_dir, storage_options=storage_options, fs=fs)
    return pm.all_pipelines(show=show)


def delete_pipeline(
    name: str,
    base_dir: str | None = None,
    remove_module_file: bool = False,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    """
    Delete a pipeline.

    Args:
        name (str): The name of the pipeline to delete.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        remove_module_file (bool, optional): Whether to delete the pipeline module. Defaults to False.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
    """
    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    p.delete(remove_module_file=remove_module_file)


def get_pipeline_summary(
    name: str | None = None,
    base_dir: str | None = None,
    config: bool = True,
    module: bool = True,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    """
    Get a summary of the pipelines.

    Args:
        name (str): The name of the pipeline.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        config (bool, optional): Whether to show the configuration. Defaults to True.
        module (bool, optional): Whether to show the module. Defaults to True.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
    Returns:
        dict: A dictionary containing the pipeline summary.
    """
    p = PipelineManager(base_dir=base_dir, storage_options=storage_options, fs=fs)
    summary = p.get_summary(name=name, config=config, module=module)

    return summary


def new_pipeline(
    name: str,
    overwrite: bool = False,
    pipeline_config: dict | PipelineConfig = {},
    params: dict = {},
    run: dict | PipelineRunConfig = {},
    schedule: dict | PipelineScheduleConfig = {},
    tracker: dict | PipelineTrackerConfig = {},
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    """
    Add a new pipeline.

    Args:
        name (str): The name of the pipeline.
        overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        pipeline_config (dict | PipelineConfig, optional): The pipeline configuration. Defaults to {}.
        params (dict, optional): The function configuration. Defaults to {}.
        run (dict | PipelineRunConfig, optional): The run configuration. Defaults to {}.
        schedule (dict | PipelineScheduleConfig, optional): The schedule configuration. Defaults to {}.
        tracker (dict | PipelineTrackerConfig, optional): The tracker configuration. Defaults to {}.
        base_dir (str | None, optional): The base directory of the pipeline. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
    """
    add_pipeline(
        name=name,
        overwrite=overwrite,
        pipeline_config=pipeline_config,
        params=params,
        run=run,
        schedule=schedule,
        tracker=tracker,
        base_dir=base_dir,
        storage_options=storage_options,
        fs=fs,
    )


def run_pipeline(
    name: str,
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    reload: bool = False,
    **kwargs,
) -> Any:
    """
    Run the pipeline with the given parameters.

    Args:
        name (str): The name of the pipeline.
        executor (str | None, optional): The executor to use. Defaults to None.
        inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
        with_tracker (bool, optional): Whether to use the tracker. Defaults to False.
        with_opentelemetry (bool, optional): Whether to use OpenTelemetry. Defaults to False.
        base_dir (str | None, optional): The base path for the pipeline. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
        reload (bool, optional): Whether to reload the pipeline. Defaults to False.
        **kwargs: Additional keyword arguments.
    Returns:
        Any: The result of running the pipeline.
    """
    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    return p.run(
        executor=executor,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        reload=reload,
        **kwargs,
    )


def run_pipeline_job(
    name: str,
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool | None = None,
    with_opentelemetry: bool | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    reload: bool = False,
    **kwargs,
) -> Any:
    """
    Add a job to run the pipeline with the given parameters to the worker.
    Executes the job immediatly and returns the job result.

    Args:
        name (str): The name of the job.
        executor (str | None, optional): The executor to use for the job. Defaults to None.
        inputs (dict | None, optional): The inputs for the job. Defaults to None.
        final_vars (list | None, optional): The final variables for the job. Defaults to None.
        with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
        with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the job. Defaults to None.
        base_dir (str | None, optional): The base path for the job. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
        reload (bool, optional): Whether to reload the job. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        Any: The result of running the job.
    """

    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    return p.run_job(
        executor=executor,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        reload=reload,
        **kwargs,
    )


def start_mqtt_listener(
    name: str,
    topic: str | None = None,
    host: str = "localhost",
    port: int = 1883,
    user: str | None = None,
    pw: str | None = None,
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool | None = None,
    with_opentelemetry: bool | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    reload: bool = False,
    result_expiration_time: float | dt.timedelta = 0,
    as_job: bool = False,
    background: bool = False,
    **kwargs,
) -> MQTTClient:
    """
    Run a pipeline when a message is received on a given topic.

    Args:
        name (str): The name of the pipeline.
        topic (str | None, optional): The topic to subscribe to. Defaults to None.
        host (str, optional): The host of the MQTT broker. Defaults to "localhost".
        port (int, optional): The port of the MQTT broker. Defaults to 1883.
        user (str | None, optional): The username for the MQTT broker. Defaults to None.
        pw (str | None, optional): The password for the MQTT broker. Defaults to None.
        inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
        executor (str | None, optional): The executor to use for the pipeline. Defaults to None.
        with_tracker (bool | None, optional): Whether to use a tracker for the pipeline. Defaults to None.
        with_opentelemetry (bool | None, optional): Whether to use OpenTelemetry for the pipeline. Defaults to None.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
        reload (bool, optional): Whether to reload the pipeline. Defaults to False.
        result_expiration_time (float | dt.timedelta, optional): The result expiration time for the job. Defaults to 0.
        as_job (bool, optional): Whether to run the pipeline as a job. Defaults to False.
        background (bool, optional): Whether to run the pipeline in the background. Defaults to False.
        **kwargs: Additional keyword arguments.
    """
    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    p.start_mqtt_listener(
        topic=topic,
        host=host,
        port=port,
        user=user,
        pw=pw,
        inputs=inputs,
        final_vars=final_vars,
        executor=executor,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        reload=reload,
        result_expiration_time=result_expiration_time,
        as_job=as_job,
        background=background,
        **kwargs,
    )
    return p.mqtt_client[name]


def schedule_pipeline(
    name: str,
    trigger_type: str = "cron",
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    paused: bool = False,
    coalesce: str = "latest",
    misfire_grace_time: float | dt.timedelta | None = None,
    max_jitter: float | dt.timedelta | None = None,
    max_running_jobs: int | None = None,
    conflict_policy: str = "do_nothing",
    # job_result_expiration_time: float | dt.timedelta = 0,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> str:
    """
    Schedule a pipeline for execution.

    Args:
        name (str): The name of the pipeline.
        executor (str | None, optional): The executor to use for running the pipeline. Defaults to None.
        trigger_type (str, optional): The type of schedule. Defaults to "cron".
        auto_start (bool, optional): Whether to automatically start the pipeline. Defaults to True.
        background (bool, optional): Whether to run the pipeline in the background. Defaults to False.
        inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the pipeline. Defaults to False.
        with_opentelemetry (bool, optional): Whether to use OpenTelemetry for the pipeline. Defaults to False.
        paused (bool, optional): Whether to start the pipeline in a paused state. Defaults to False.
        coalesce (str, optional): The coalesce strategy for the pipeline. Defaults to "latest".
        misfire_grace_time (float | dt.timedelta | None, optional): The grace time for misfired jobs. Defaults to None.
        max_jitter (float | dt.timedelta | None, optional): The maximum number of seconds to randomly add to the
            scheduled. Defaults to None.
        max_running_jobs (int | None, optional): The maximum number of running jobs. Defaults to None.
        conflict_policy (str, optional): The conflict policy for the pipeline. Defaults to "do_nothing".
        job_result_expiration_time (float | dt.timedelta | None, optional): The result expiration time for the job.
                Defaults to None.
        base_dir (str | None, optional): The base path for the pipeline. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        str: The ID of the scheduled pipeline.
    """
    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    return p.schedule(
        executor=executor,
        tigger_type=trigger_type,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        paused=paused,
        coalesce=coalesce,
        misfire_grace_time=misfire_grace_time,
        max_jitter=max_jitter,
        max_running_jobs=max_running_jobs,
        # job_result_expiration_time=job_result_expiration_time,
        conflict_policy=conflict_policy,
        **kwargs,
    )


def show_pipeline_dag(
    name: str,
    format: str = "png",
    show: bool = False,
    save: bool = False,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    reload: bool = False,
):
    """
    Display the pipeline with the given name.

    Parameters:
        name (str): The name of the pipeline.
        format (str, optional): The format of the displayed pipeline. Defaults to "png".
        view (bool, optional): Whether to display the pipeline. Defaults to False.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
        reload (bool, optional): Whether to reload the pipeline. Defaults to False.
    """
    p = Pipeline(name=name, base_dir=base_dir, storage_options=storage_options, fs=fs)
    p.show_dag(format=format, show=show, reload=reload, save=save)


def show_pipeline_summary(
    name: str | None = None,
    base_dir: str | None = None,
    config: bool = True,
    module: bool = True,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    """
    Show a summary of the pipelines.

    Args:
        name (str): The name of the pipeline.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        config (bool, optional): Whether to show the configuration. Defaults to True.
        module (bool, optional): Whether to show the module. Defaults to True.
        storage_options (dict, optional): The fsspec storage options. Defaults to {}.
        fs (AbstractFileSystem | None, optional): The fsspec filesystem to use. Defaults to None.
    """
    p = PipelineManager(base_dir=base_dir, storage_options=storage_options, fs=fs)
    p.show_summary(name=name, config=config, module=module)
